"""
Implements the Google Cloud DNS provider.
This API is quite complicated to use, as it used some unique
concepts compared to other providers.
First of all, it uses a full-fledged OAuth2 authentication,
involving signing a JWT and retrieve a Bearer token.
This hard work is done in the authenticate() process,
using the strong and well known "cryptography" package.
Second, Google Cloud DNS API contains this really particular patterns:
  - all records of the same type and name are stacked together in a RecordSet representation,
    which contains in the rrdatas array all current values for this type/name pair, including
    explicitly monovalued entries like A or CNAME.
  - modifications can only done through a create/delete pattern: no way to update a record
  - more importantly, this approach extends to all values of a given type/name pair: it means
    that adding/removing a new value to a TXT entry requires to delete all values of this entry,
    then recreate it with all values desired (the old ones plus the new one for adding, the old
    ones minus the removed one for removing)
So all the hard work in this provider, apart from the authentication process, is to convert the
Lexicon monovalued entries representation to/from the Google multivalued and stacked representation
through create/update/list/delete processes.
"""

import hashlib
import json
import logging
import time
from argparse import ArgumentParser
from base64 import b64decode, urlsafe_b64encode
from typing import List

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    Provider class for Google Cloud DNS

    We need serveral parameters, which are available in the JSON file provided
    by Google when associating a private key to the relevant service account.
    So this JSON file is the natural input to configure the provider.
    It can be provided as a path to the JSON file, or as its content encoded
    in base64, which is a suitable portable way in particular for Docker containers.
    In both cases the content is loaded as bytes, on loaded in a private instance variable.
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["googledomains.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            The Google Cloud DNS provider requires the JSON file which contains the service account info to connect to the API.
            This service account must own the project role DNS > DNS administrator for the project associated to the DNS zone.
            You can create a new service account, associate a private key, and download its info through this url:
            https://console.cloud.google.com/iam-admin/serviceaccounts?authuser=2"""
        parser.add_argument(
            "--auth-service-account-info",
            help="""
        specify the service account info in the Google JSON format:
        can be either the path of a file prefixed by 'file::' (eg. file::/tmp/service_account_info.json)
        or the base64 encoded content of this file prefixed by 'base64::'
        (eg. base64::eyJhbGciOyJ...)""",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._token = None

        if self._get_provider_option("auth_service_account_info").startswith("file::"):
            with open(
                self._get_provider_option("auth_service_account_info").replace(
                    "file::", ""
                ),
                "rb",
            ) as file:
                service_account_info_bytes = file.read()
        elif self._get_provider_option("auth_service_account_info").startswith(
            "base64::"
        ):
            service_account_info_bytes = b64decode(
                self._get_provider_option("auth_service_account_info").replace(
                    "base64::", ""
                )
            )
        else:
            raise Exception(
                "Invalid value for --auth-service-account-info, should be a path "
                "prefixed with 'file::' or a base64 value prefixed by 'base64::'."
            )

        self._service_account_info = json.loads(
            service_account_info_bytes.decode("utf-8")
        )

        if (
            not self._service_account_info["client_email"]
            or not self._service_account_info["private_key"]
            or not self._service_account_info["project_id"]
        ):
            raise Exception(
                "Invalid service account info (missing either client_email/private_"
                "key/project_id key)."
            )

    # the complete list of zones may be paginated. So we recursively call
    # the list managedZones api until no page_token is given, keeping a list
    # of matched zone ids as we go.
    def _get_managed_zone_ids(self, zone_ids, page_token=None):
        results = self._get("/managedZones", {"pageToken": page_token})
        zone_ids += [
            managedZone["id"]
            for managedZone in results["managedZones"]
            if managedZone["dnsName"] == f"{self.domain}."
        ]

        if "nextPageToken" in results:
            return self._get_managed_zone_ids(zone_ids, results["nextPageToken"])
        return zone_ids

    # We have a real authentication here, that uses the OAuth protocol:
    #   - a JWT token is forged with the Google Cloud DNS access claims,
    #     using the service account info loaded by the constructor,
    #   - this JWT token is signed by a PKCS1v15 signature using
    #     the RSA private key associated to the service account,
    #   - this JWT token is then submitted to the Google API, which returns an access token
    #   - this access token will be used for every future HTTP request
    #     to the Google Cloud DNS API to authenticate the user.
    #   - finally we make a first authenticated request to retrieve
    #     the managed zone id, which will also be used on future requests.
    # This access token has a default lifetime of 10 minutes,
    # but is used only for the current Lexicon operation, so it should be sufficient.
    def authenticate(self):
        jwt_header_bytes = urlsafe_b64encode(
            json.dumps({"alg": "RS256", "typ": "JWT"}).encode("utf-8")
        )

        epoch_time = int(time.time())
        jwt_claims_bytes = urlsafe_b64encode(
            json.dumps(
                {
                    "iss": self._service_account_info["client_email"],
                    "scope": "https://www.googleapis.com/auth/ndev.clouddns.readwrite",
                    "aud": "https://www.googleapis.com/oauth2/v4/token",
                    "exp": epoch_time + 60 * 10,
                    "iat": epoch_time,
                }
            ).encode("utf-8")
        )

        private_key = serialization.load_pem_private_key(
            self._service_account_info["private_key"].encode("utf-8"),
            password=None,
            backend=default_backend(),
        )
        jwt_sign_bytes = urlsafe_b64encode(
            private_key.sign(
                b".".join([jwt_header_bytes, jwt_claims_bytes]),
                padding.PKCS1v15(),
                hashes.SHA256(),
            )
        )

        jwt_bytes = b".".join([jwt_header_bytes, jwt_claims_bytes, jwt_sign_bytes])

        auth_request = requests.request(
            "POST",
            "https://www.googleapis.com/oauth2/v4/token",
            data={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": jwt_bytes,
            },
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

        auth_request.raise_for_status()
        post_result = auth_request.json()

        if not post_result["access_token"]:
            raise AuthenticationError(
                "Error, could not grant RW access on the "
                f"Google Cloud DNS API for user: {self._get_provider_option('auth_email')}"
            )

        self._token = post_result["access_token"]

        targeted_managed_zone_ids = self._get_managed_zone_ids([])
        if not targeted_managed_zone_ids:
            raise AuthenticationError(
                f"Error, domain {self.domain} is not registered for this project"
            )

        self.domain_id = targeted_managed_zone_ids[0]

    def cleanup(self) -> None:
        pass

    # List all records for the given type/name/content.
    # It is quite straight forward to request data, the biggest operation is to convert
    # the stacked multivalued RecordSets into Lexicon monovalued entries.
    # Please note that we could provide type and name to the API to make the filtering,
    # but providing the type makes the name mandatory with the Google Cloud DNS API, and
    # name is not always available (we can ask for every TXT record for example). So to stick to
    # the most general case, its preferable to always get all records and be free to filter
    # the way we want afterwards.
    def list_records(self, rtype=None, name=None, content=None):
        results = self._get(f"/managedZones/{self.domain_id}/rrsets")

        records = []

        for rrset in results["rrsets"]:
            for rrdata in rrset["rrdatas"]:
                record = {
                    "type": rrset["type"],
                    "name": self._full_name(rrset["name"]),
                    "ttl": rrset["ttl"],
                    "content": rrdata,
                }
                self._clean_TXT_record(record)
                record["id"] = Provider._identifier(record)
                records.append(record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    # Create the record with provided type, name and content.
    # Because of the way this API is constructed, it is quite complex in fact.
    # Indeed we need to know if there is already a RecordSet for the type/name pair, and update
    # or create accordingly the RecordSet. Furthermore, we need first to delete the old RecordSet
    # if it exists, to replace it with the RecordSet containing the new content we want.
    def create_record(self, rtype, name, content):
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        identifier = Provider._identifier(
            {"type": rtype, "name": self._full_name(name), "content": content}
        )

        query_params = {"type": rtype, "name": self._fqdn_name(name)}

        results = self._get(
            f"/managedZones/{self.domain_id}/rrsets", query_params=query_params
        )

        rrdatas = []
        changes = {}
        if results["rrsets"]:
            rrset = results["rrsets"][0]
            for rrdata in rrset["rrdatas"]:
                if rrdata == Provider._normalize_content(rrset["type"], content):
                    LOGGER.debug("create_record (ignored, duplicate): %s", identifier)
                    return True

            changes["deletions"] = [
                {
                    "name": rrset["name"],
                    "type": rrset["type"],
                    "ttl": rrset["ttl"],
                    "rrdatas": rrset["rrdatas"][:],
                }
            ]

            rrdatas = rrset["rrdatas"][:]

        rrdatas.append(Provider._normalize_content(rtype, content))

        changes["additions"] = [
            {
                "name": self._fqdn_name(name),
                "type": rtype,
                "ttl": self._get_lexicon_option("ttl"),
                "rrdatas": rrdatas,
            }
        ]

        self._post(f"/managedZones/{self.domain_id}/changes", data=changes)

        LOGGER.debug("create_record: %s", identifier)

        return True

    # Update a record for the given identifier or type/name pair
    # with the given content if provided.
    # Again because of the API specification, updating is even more complex than creating,
    # as we need to take into account every RecordSet that should be destroyed then recreated.
    # As all the hard work has been done on list_record, create_record and delete_record, we use a
    # combination of these three methods to obtain the state we want.
    # Even if this make the operation very costly regarding the number of requests to do, it allows
    # the implementation to be way more readable (without that, it would take grossly the size of
    # he three quoted methods).
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier and (not rtype or not name):
            raise Exception("Error, identifier or rtype+name parameters are required.")

        if identifier:
            records = self.list_records()
            records_to_update = [
                record for record in records if record["id"] == identifier
            ]
        else:
            records_to_update = self.list_records(rtype=rtype, name=name)

        if not records_to_update:
            raise Exception(
                f"Error, could not find a record for given identifier: {identifier}"
            )

        if len(records_to_update) > 1:
            LOGGER.warning(
                "Warning, multiple records found for given parameters, "
                "only first one will be updated: %s",
                records_to_update,
            )

        record_identifier = records_to_update[0]["id"]

        original_level = LOGGER.getEffectiveLevel()
        LOGGER.setLevel(logging.WARNING)
        self.delete_record(record_identifier)

        new_record = {
            "type": rtype if rtype else records_to_update[0]["type"],
            "name": name if name else records_to_update[0]["name"],
            "content": content if content else records_to_update[0]["content"],
        }

        self.create_record(
            new_record["type"], new_record["name"], new_record["content"]
        )
        LOGGER.setLevel(original_level)

        LOGGER.debug(
            "update_record: %s => %s",
            record_identifier,
            Provider._identifier(new_record),
        )

        return True

    # Delete a record for the given identifier or the given type/name/content.
    # Really complex to do, because a lot of RecordSets can be updated (so destroyed and recreated)
    # depending on the given condition (eg. with content alone, every record could be inspected).
    # There is mainly to cases:
    #   - either an association of one or more between type, name and content is given
    #   - either an identifier is given, and we extract the type + name + content
    #     to process as the first case.
    # Anyway, we will need to parse every RecordSet available, and for each of them
    # that matches the conditions:
    #   - mark as deletion the existing RecordSet
    #   - remove the targeted content from the RecordSet
    #   - mark as addition the update RecordSet with the subset of rrdatas if rrdatas is not empty
    #   - do not mark as additions RecordSets whose rrdatas subset become empty:
    #     for this type/name pair, all RecordSet needs to go away.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        results = self._get(f"/managedZones/{self.domain_id}/rrsets")

        if identifier:
            changes = self._process_records_to_delete_by_identifier(results, identifier)
        else:
            changes = self._process_records_to_delete_by_parameters(
                results, rtype, name, content
            )

        if not changes:
            raise Exception(
                "Could not find existing record matching the given parameters."
            )

        self._post(f"/managedZones/{self.domain_id}/changes", data=changes)

        LOGGER.debug("delete_records: %s %s %s %s", identifier, rtype, name, content)

        return True

    # Calculate the changes to do based on the record to remove identified by its identifier.
    # This implementation find the corresponding record, and use its type + name + value to
    # delegate the processing to _process_records_to_delete_by_parameters.
    def _process_records_to_delete_by_identifier(self, results, identifier):
        for rrset in results["rrsets"]:
            for rrdata in rrset["rrdatas"]:
                record = {
                    "type": rrset["type"],
                    "name": self._full_name(rrset["name"]),
                    "content": rrdata,
                }

                self._clean_TXT_record(record)
                record_identifier = Provider._identifier(record)

                if identifier == record_identifier:
                    return self._process_records_to_delete_by_parameters(
                        results, record["type"], record["name"], record["content"]
                    )

        return None

    # Calculate the changes to do based on the records to remove identified by type/name/content.
    # Additions and deletions are registered accordingly in the changes, and RecordSet with empty
    # rrdatas after its subset are not marked in additions to be completely removed
    # from the DNS zone.
    def _process_records_to_delete_by_parameters(
        self, results, rtype=None, name=None, content=None
    ):
        rrsets_to_modify = results["rrsets"]

        if rtype:
            rrsets_to_modify = [
                rrset for rrset in rrsets_to_modify if rrset["type"] == rtype
            ]
        if name:
            rrsets_to_modify = [
                rrset
                for rrset in rrsets_to_modify
                if rrset["name"] == self._fqdn_name(name)
            ]
        if content:
            rrsets_to_modify = [
                rrset
                for rrset in rrsets_to_modify
                if (f'"{content}"' if rrset["type"] == "TXT" else content)
                in rrset["rrdatas"]
            ]

        changes = {"additions": [], "deletions": []}

        for rrset_to_modify in rrsets_to_modify:
            changes["deletions"].append(
                {
                    "name": rrset_to_modify["name"],
                    "type": rrset_to_modify["type"],
                    "ttl": rrset_to_modify["ttl"],
                    "rrdatas": rrset_to_modify["rrdatas"][:],
                }
            )

            if content:
                new_rrdatas = rrset_to_modify["rrdatas"][:]
                new_rrdatas.remove(
                    f'"{content}"' if rrset_to_modify["type"] == "TXT" else content
                )
                if new_rrdatas:
                    changes["additions"].append(
                        {
                            "name": rrset_to_modify["name"],
                            "type": rrset_to_modify["type"],
                            "ttl": rrset_to_modify["ttl"],
                            "rrdatas": new_rrdatas,
                        }
                    )

        if not changes["additions"] and not changes["deletions"]:
            return None

        return changes

    # With Google Cloud DNS API, content of CNAME entries must be FQDN (with a trailing dot),
    #   and content of TXT entries must be quoted. This static method ensures that.
    @staticmethod
    def _normalize_content(rtype, content):
        if rtype == "TXT":
            return f'"{content}"'
        if rtype == "CNAME":
            return f"{content}." if not content.endswith(".") else content

        return content

    # Google Cloud DNS API does not provide identifier for RecordSets.
    # So we need to calculate our own identifier at runtime.
    # It is based on a SHA256 hash with the most relevant
    # parameters of a record: type, name and content.
    # Note that the identifier is calculated on a Lexicon monovalued entry, not a Google stacked
    # multivalued RecordSet, to make it usable during Lexicon calls to updates and deletions.
    @staticmethod
    def _identifier(record):
        m = hashlib.sha256()
        m.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        m.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
        m.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

        return m.hexdigest()[0:7]

    # The request, when authenticated, is really standard:
    #   - the request body is encoded as application/json for POST
    #     (so the use of 'json' config instead of 'data' in request),
    #   - the body response is also encoded as application/json for GET and POST,
    #   - and the request headers must contain the access token in the 'Authorization' field.
    def _request(self, action="GET", url="/", data=None, query_params=None):
        request = requests.request(
            action,
            f"https://content.googleapis.com/dns/v1/projects/{self._service_account_info['project_id']}{url}",
            params=None if not query_params else query_params,
            json=None if not data else data,
            headers={"Authorization": f"Bearer {self._token}"},
        )

        request.raise_for_status()
        return request.json()



#!/usr/bin/env python
"""Module for Lexicon command-line interface"""
import json
import logging
import os
import sys
import warnings
from typing import Dict, List, Optional, Union

from lexicon._private.parser import generate_cli_main_parser
from lexicon.client import Client
from lexicon.config import ConfigResolver

logger = logging.getLogger(__name__)


def generate_list_table_result(
    lexicon_logger: logging.Logger,
    output: Optional[Union[bool, List[Dict]]] = None,
    without_header: Optional[bool] = None,
) -> Optional[str]:
    """Convert returned data from list actions into a nice table for command line usage"""
    if not isinstance(output, list):
        lexicon_logger.debug(
            "Command output is not a list, and then cannot "
            "be printed with --quiet parameter not enabled."
        )
        return None

    array = [
        [
            row.get("id", ""),
            row.get("type", ""),
            row.get("name", ""),
            row.get("content", ""),
            row.get("ttl", ""),
        ]
        for row in output
    ]

    # Insert header (insert before calculating the max width of each column
    # to take headers size into account)
    if not without_header:
        headers = ["ID", "TYPE", "NAME", "CONTENT", "TTL"]
        array.insert(0, headers)

    column_widths = [0, 0, 0, 0, 0]
    # Find max width for each column
    for row in array:
        for idx, col in enumerate(row):
            width = len(str(col))
            if width > column_widths[idx]:
                column_widths[idx] = width

    # Add a 'nice' separator
    if not without_header:
        array.insert(1, ["-" * column_widths[idx] for idx in range(len(column_widths))])

    # Construct table to be printed
    table = []
    for row in array:
        row_list = []
        for idx, col in enumerate(row):
            row_list.append(str(col).ljust(column_widths[idx]))
        table.append(" ".join(row_list))

    # Return table
    return os.linesep.join(table)


def generate_table_results(
    output: Optional[Union[bool, List[Dict]]] = None,
    without_header: Optional[bool] = None,
) -> str:
    """Convert returned data from non-list actions into a nice table for command line usage"""
    array = []
    str_output = str(output)

    if not without_header:
        array.append("RESULT")
        array.append("-" * max(6, len(str_output)))

    array.append(str_output)
    return os.linesep.join(array)


def handle_output(
    results: Union[bool, List[Dict]], output_type: str, action: str
) -> None:
    """Print the relevant output for given output_type"""
    if output_type == "QUIET":
        return

    if not output_type == "JSON":
        if action == "list":
            table = generate_list_table_result(
                logger, results, output_type == "TABLE-NO-HEADER"
            )
        else:
            table = generate_table_results(results, output_type == "TABLE-NO-HEADER")
        if table:
            print(table)
    else:
        try:
            json_str = json.dumps(results)
            if json_str:
                print(json_str)
        except TypeError:
            logger.debug(
                "Output is not JSON serializable, and then cannot "
                "be printed with --output=JSON parameter."
            )


def main() -> None:
    """Main function of Lexicon."""
    # Dynamically determine all the providers available and gather command line arguments.
    parsed_args = generate_cli_main_parser().parse_args()

    log_level = logging.getLevelName(parsed_args.log_level)
    logging.basicConfig(stream=sys.stdout, level=log_level, format="%(message)s")
    logger.debug("Arguments: %s", parsed_args)

    # In the CLI context, will get configuration interactively:
    #   * from the command line
    #   * from the environment variables
    #   * from lexicon configuration files found in given --config-dir (default is current dir)
    config = ConfigResolver()
    config.with_args(parsed_args).with_env().with_config_dir(parsed_args.config_dir)

    client = Client(config)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        results = client.execute()

    action = config.resolve("lexicon:action")
    if not action:
        raise ValueError("Parameter action is not set.")

    handle_output(results, parsed_args.output, action)


if __name__ == "__main__":
    main()



"""
Lexicon DevNomads Provider

Author: Loek Geleijn, 2025

API Docs: https://api.devnomads.nl/api/documentation#/Dns

Implementation notes:
The DevNomads API basically implements the PowerDNS API but applies client
specific authentication and filtering.
"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class DevNomadsProviderError(Exception):
    """Generic DevNomads exception"""


class Provider(BaseProvider):
    """Provider class for DevNomads"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="Specify token for authentication.")

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = "https://api.devnomads.nl/services/dns"

        if self.api_endpoint.endswith("/"):
            self.api_endpoint = self.api_endpoint[:-1]

        self.auth_token = self._get_provider_option("auth_token")
        if not self.auth_token:
            raise DevNomadsProviderError(
                "DevNomads auth token not defined (auth_token)"
            )

        self._zone_data = None

    def zone_data(self):
        """Get zone data"""
        if self._zone_data is None:
            self._zone_data = self._get(
                "/zones/" + self._ensure_dot(self.domain)
            ).json()
        return self._zone_data

    def authenticate(self):
        self.zone_data()
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def _make_identifier(self, rtype, name, content):
        return f"{rtype}/{name}={content}"

    def _parse_identifier(self, identifier):
        parts = identifier.split("/")
        rtype = parts[0]
        parts = parts[1].split("=")
        name = parts[0]
        content = "=".join(parts[1:])
        return rtype, name, content

    def list_records(self, rtype=None, name=None, content=None):
        records = []
        for rrset in self.zone_data()["rrsets"]:
            if (
                name is None or self._fqdn_name(rrset["name"]) == self._fqdn_name(name)
            ) and (rtype is None or rrset["type"] == rtype):
                for record in rrset["records"]:
                    if content is None or record["content"] == self._clean_content(
                        rtype, content
                    ):
                        records.append(
                            {
                                "type": rrset["type"],
                                "name": self._full_name(rrset["name"]),
                                "ttl": rrset["ttl"],
                                "content": self._unclean_content(
                                    rrset["type"], record["content"]
                                ),
                                "id": self._make_identifier(
                                    rrset["type"], rrset["name"], record["content"]
                                ),
                            }
                        )
        LOGGER.debug(f"list_records: {records}")
        return records

    def _clean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        elif rtype == "CNAME":
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            content = content.strip('"')
        elif rtype == "CNAME":
            content = self._full_name(content)
        return content

    def create_record(self, rtype, name, content):
        rname = self._fqdn_name(name)
        newcontent = self._clean_content(rtype, content)

        updated_data = {
            "name": rname,
            "type": rtype,
            "records": [],
            "ttl": self._get_lexicon_option("ttl") or 600,
            "changetype": "REPLACE",
        }

        updated_data["records"].append({"content": newcontent, "disabled": False})

        for rrset in self.zone_data()["rrsets"]:
            if rrset["name"] == rname and rrset["type"] == rtype:
                updated_data["ttl"] = rrset["ttl"]

                for record in rrset["records"]:
                    if record["content"] != newcontent:
                        updated_data["records"].append(
                            {
                                "content": record["content"],
                                "disabled": record["disabled"],
                            }
                        )
                break

        request = {"rrsets": [updated_data]}
        LOGGER.debug(f"request: {requests}")

        self._patch("/zones/" + self._ensure_dot(self.domain), data=request)
        self._zone_data = None
        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is not None:
            rtype, name, content = self._parse_identifier(identifier)

        LOGGER.debug(f"delete {rtype} {name} {content}")
        if rtype is None or name is None:
            raise Exception("Must specify at least both rtype and name")

        for rrset in self.zone_data()["rrsets"]:
            if rrset["type"] == rtype and self._fqdn_name(
                rrset["name"]
            ) == self._fqdn_name(name):
                update_data = rrset

                if "comments" in update_data:
                    del update_data["comments"]

                if content is None:
                    update_data["records"] = []
                    update_data["changetype"] = "DELETE"
                else:
                    new_record_list = []
                    for record in update_data["records"]:
                        if (
                            self._clean_content(rrset["type"], content)
                            != record["content"]
                        ):
                            new_record_list.append(record)

                    update_data["records"] = new_record_list
                    update_data["changetype"] = "REPLACE"
                break

        request = {"rrsets": [update_data]}
        LOGGER.debug(f"request: {request}")

        self._patch("/zones/" + self._ensure_dot(self.domain), data=request)

        self._zone_data = None
        return True

    def update_record(self, identifier, rtype=None, name=None, content=None):
        self.delete_record(identifier)
        return self.create_record(rtype, name, content)

    def _patch(self, url="/", data=None, query_params=None):
        return self._request("PATCH", url, data=data, query_params=query_params)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Authorization": f"Bearer {self.auth_token}",
                "Content-Type": "application/json",
            },
        )
        LOGGER.debug(f"response: {response.text}")
        response.raise_for_status()
        return response

    @classmethod
    def _ensure_dot(cls, text):
        """
        This function makes sure a string contains a dot at the end
        """
        if text.endswith("."):
            return text
        return text + "."



"""Module provider for Scaleway
Documentation about Domains and DNS APIs for Scaleway can be found in
https://www.scaleway.com/en/developers/api/domains-and-dns
Documentation to obtain an API key can be found in
https://www.scaleway.com/en/developers/api/
"""

import json
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.interfaces import Provider as BaseProvider


class Provider(BaseProvider):
    """
    Provide Scaleway Domains and DNS API Implementation for lexicon.
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["scw.cloud"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-secret-key",
            help="specify Scaleway API key",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.default_ttl = 3600
        self.endpoint = "https://api.scaleway.com/domain/v2beta1"

    def authenticate(self):
        self.domain_id = self.domain.lower()
        self._get(self._records_url())

    def create_record(self, rtype, name, content):
        records = self.list_records(rtype, name, content)
        if records:
            return True
        patch = {
            "changes": [
                {
                    "add": {
                        "records": [
                            {
                                "name": name,
                                "data": content,
                                "type": rtype,
                                "ttl": self._get_lexicon_option("ttl")
                                or self.default_ttl,
                            },
                        ],
                    },
                },
            ],
        }
        self._patch(self._records_url(), patch)
        return True

    def list_records(self, rtype=None, name=None, content=None):
        # Up to 20 records are listed with default page-size, we need to request all pages.
        page = 1
        total = 0
        records = []
        while page == 1 or len(records) < total:
            results = self._get(self._records_url() + f"?page={page}")
            for result in results["records"]:
                record = {
                    "id": result["id"],
                    "type": result["type"],
                    "name": self._full_name(result["name"]),
                    "ttl": result["ttl"],
                    "content": self._decode_content(result["type"], result["data"]),
                }
                records.append(record)

            total = results["total_count"]
            page += 1
        return self._filter_records(records, rtype, name, content)

    @staticmethod
    def _decode_content(rtype, data):
        if not data or rtype != "TXT":
            return data
        if data[0] == '"' and data[len(data) - 1] == '"':
            return (
                data[1:-1].replace('" "', "").replace('\\"', '"').replace("\\\\", "\\")
            )
        return data

    def _filter_records(self, records, rtype=None, name=None, content=None):
        if not rtype and not name and not content:
            return records

        filtered = []
        for record in records:
            if rtype and record["type"] != rtype:
                continue
            if name and record["name"] != self._full_name(name):
                continue
            if content and record["content"] != content:
                continue
            filtered.append(record)
        return filtered

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        changes = []
        if identifier:
            changes.append(
                {
                    "delete": {
                        "id": identifier,
                    },
                }
            )
        else:
            records = self.list_records(rtype, name, content)
            if not records:
                return True
            for record in records:
                changes.append(
                    {
                        "delete": {
                            "id": record["id"],
                        },
                    }
                )
        patch = {
            "changes": changes,
        }
        self._patch(self._records_url(), patch)
        return True

    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            records = self.list_records(None, name)
            if not records:
                raise Exception(f"Record {name} not found with type {rtype}")
            if len(records) > 1:
                raise Exception(f"More than one {rtype} record found for {name}")
            identifier = records[0]["id"]
        patch = {
            "changes": [
                {
                    "set": {
                        "id": identifier,
                        "records": [
                            {
                                "name": name,
                                "data": content,
                                "type": rtype,
                                "ttl": self._get_lexicon_option("ttl")
                                or self.default_ttl,
                            },
                        ],
                    },
                },
            ],
        }
        print(patch)
        self._patch(self._records_url(), patch)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Auth-Token": self._get_provider_option("auth_secret_key"),
        }
        response = requests.request(
            action,
            self.endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=headers,
        )
        response.raise_for_status()
        return response.json()

    def _records_url(self):
        return f"/dns-zones/{self.domain_id}/records"



"""Module provider for Infomaniak"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

ENDPOINT = "https://api.infomaniak.com"


class Provider(BaseProvider):
    """Provider class for Infomaniak"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["infomaniak.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            Infomaniak Provider requires a token with domain scope.
            It can be generated for your Infomaniak account on the following URL:
            https://manager.infomaniak.com/v3/infomaniak-api"""

        parser.add_argument("--auth-token", help="specify the token")

    def __init__(self, config):
        super(Provider, self).__init__(config)

        # Handling missing required parameters
        if not self._get_provider_option("auth_token"):
            raise Exception("Error, token is not defined")

        # Construct DNS Infomaniak environment
        self.domain_id = None
        self.endpoint_api = ENDPOINT
        self.session = None

    def authenticate(self):
        domains = self._get(
            "/1/product", {"service_name": "domain", "customer_name": self.domain}
        )

        LOGGER.debug("found domains %s", domains)

        for domain in domains["data"]:
            if domain["customer_name"] == self.domain:
                self.domain_id = domain["id"]
                break
        else:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        ttl = self._get_lexicon_option("ttl")

        records = list(
            filter(
                lambda x: x["content"] == content
                and x["type"] == rtype
                and self._relative_name(x["name"]) == self._relative_name(name),
                self.list_records(rtype, name, content),
            )
        )

        if len(records) > 0:
            LOGGER.debug(
                "create_record (ignored, duplicate): %s %s %s", rtype, name, content
            )
            return True

        data = {
            "type": rtype,
            "source": self._relative_name(name),
            "target": content,
        }

        if ttl:
            data["ttl"] = ttl

        result = self._post(f"/1/domain/{self.domain_id}/dns/record", data)

        LOGGER.debug("create_record: %s", result["data"])

        return True

    def list_records(self, rtype=None, name=None, content=None):
        records = []

        record_data = self._get(f"/1/domain/{self.domain_id}/dns/record")

        for record in record_data["data"]:
            records.append(
                {
                    "type": record["type"],
                    "name": record["source_idn"],
                    "ttl": record["ttl"],
                    "content": record["target_idn"],
                    "id": record["id"],
                }
            )

        if rtype:
            records = [record for record in records if record["type"] == rtype]

        if name:
            records = [
                record
                for record in records
                if record["name"].lower() == self._full_name(name.lower())
            ]

        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)

        return records

    def _get_record(self, identifier):
        record_data = self._get(f"/1/domain/{self.domain_id}/dns/record/{identifier}")

        record = {
            "type": record_data["data"]["type"],
            "name": record_data["data"]["source_idn"],
            "ttl": record_data["data"]["ttl"],
            "content": record_data["data"]["target_idn"],
            "id": record_data["data"]["id"],
        }

        LOGGER.debug("get_record: %s", record)

        return record

    def update_record(self, identifier, rtype=None, name=None, content=None):
        records = self.list_records(rtype, name)

        if not identifier:
            if len(records) == 1:
                identifier = records[0]["id"]
                record = records[0]
            elif len(records) > 1:
                raise Exception("Several record identifiers match the request")
            else:
                raise Exception("Record identifier could not be found")
        else:
            record = self._get_record(identifier)

        data = {"ttl": record["ttl"]}
        if name:
            data["source"] = self._relative_name(name)
        if content:
            data["target"] = content

        self._put(f"/1/domain/{self.domain_id}/dns/record/{identifier}", data)

        LOGGER.debug("update_record: %s", identifier)

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/1/domain/{self.domain_id}/dns/record/{record_id}")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {}
        target = self.endpoint_api + url
        body = ""

        if data is not None:
            headers["Content-type"] = "application/json"
            body = json.dumps(data)

        headers["Authorization"] = f"Bearer {self._get_provider_option('auth_token')}"

        result = requests.request(
            method=action, url=target, params=query_params, data=body, headers=headers
        )
        result.raise_for_status()

        json_result = result.json()

        if json_result["result"] != "success":
            raise Exception("API didn't return success status")

        return json_result
